/*******************  start of original comments  ********************/
/*
 * Written by Douglas Thomson (1989/1990)
 *
 * This source code is released into the public domain.
 */

/*
 * Name:    dte - Doug's Text Editor program - window module
 * Purpose: This file contains the code associated with opening and sizing
 *           windows, and also displaying the help window.
 * File:    window.c
 * Author:  Douglas Thomson
 * System:  this file is intended to be system-independent
 * Date:    October 12, 1989
 */
/*********************  end of original comments   ********************/


/*
 * The window routines have been EXTENSIVELY rewritten.  Some routines were
 * changed so only one logical function is carried out, eg. 'initialize_window'.
 * I like the Microsoft way of resizing windows - just press the up and down
 * arrows to adjust the window to desired size.  I also like pressing one key
 * to change windows.  All of which are implemented.
 *
 * New editor name:  tde, the Thomson-Davis Editor.
 * Author:           Frank Davis
 * Date:             June 5, 1991, version 1.0
 * Date:             July 29, 1991, version 1.1
 * Date:             October 5, 1991, version 1.2
 * Date:             January 20, 1992, version 1.3
 *
 * This modification of Douglas Thomson's code is released into the
 * public domain, Frank Davis.  You may distribute it freely.
 */

#include "tdestr.h"
#include "common.h"
#include "define.h"
#include "tdefunc.h"


/*
 * Name:    initialize_window
 * Purpose: To open a new window
 * Date:    June 5, 1991
 * Returns: OK if window opened successfully
 *          ERROR if anything went wrong
 * Notes:   If this is first window then set up as normal displayed window
 *          otherwise make the present window invisible and open a new
 *          window in the same screen location as the old one.
 */
int  initialize_window( void )
{
int top;
int bottom;
windows *wp;        /* used for scanning windows */
windows *window;
register file_infos *fp;     /* used for scanning files */
register int rc;

   rc = OK;
   window = g_status.current_window;
   fp = g_status.current_file;
   if (window == NULL) {
      /*
       * special case if this is the first window on screen.
       */
      top = 0;
      bottom = g_display.nlines;
   } else {
      /*
       * else put the new window in same place as current window.
       * make current window invisible.  new window becomes current window.
       */
      top = window->top_line - 1;
      bottom = window->bottom_line;
   }

   if (create_window( &wp, top, bottom, fp ) == ERROR) {
      error( WARNING, bottom, "out of memory" );

      /*
       * This is a real nuisance. We had room for the file and the
       *  file structure, but not enough for the window as well.
       * Now we must free all the memory that has already been
       *  allocated.
       */
      if (fp->ref_count == 0) {
         if (fp->prev)
            fp->prev->next = fp->next;
         else
            g_status.file_list = fp->next;
         if (fp->next)
            fp->next->prev = fp->prev;
         g_status.end_mem = fp->start_text;
         free( fp );
         wp = g_status.current_window;
         if (wp != NULL && wp->visible)
            g_status.current_file = wp->file_info;
         else
            g_status.stop = TRUE;
      }
      rc = ERROR;
   }

   if (rc != ERROR) {
      /*
       * set up the new cursor position as appropriate
       */
      wp->cursor = cpf( fp->start_text );
      wp->ccol = wp->rcol = wp->bcol = 0;
      wp->rline = 1l;
      wp->visible = TRUE;
      wp->letter = fp->next_letter++;
      if (window != NULL)
         window->visible = FALSE;

      /*
       * the new window becomes the current window.
       */
      g_status.current_window = wp;
   }
   return( rc );
}


/*
 * Name:    next_window
 * Purpose: To move to the next visible window.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Start with current window.  If next window exists then go to
 *          it else go to the first (top) window on screen.
 */
void next_window( windows *window )
{
register int change;
register windows *wp;

   if (window != NULL) {
      change = FALSE;
      wp = g_status.window_list;
      if (wp->next != NULL) {
         while (wp != NULL && (wp->top_line-2 != window->bottom_line ||
                !wp->visible))
            wp = wp->next;
         if (wp != NULL)
            change = TRUE;
      }
      if (change != TRUE) {
         wp = g_status.window_list;
         while (wp != NULL && (wp->top_line != 1 || !wp->visible))
            wp = wp->next;
         if (wp != NULL && wp != window)
            change = TRUE;
      }
      if (change == TRUE) {
         un_copy_line( window->cursor, window, TRUE );
         g_status.current_window = wp;
         g_status.current_file = wp->file_info;
      }
   }
}


/*
 * Name:    prev_window
 * Purpose: To move to the previous visible window.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Start with current window.  If previous window exists then go to
 *          it else go to the last (bottom) window on screen.  Opposite of
 *          next_window.
 */
void prev_window( windows *window )
{
register int change;
register windows *wp;

   if (window != NULL) {
      change = FALSE;
      wp = g_status.window_list;
      if (wp->next != NULL) {
         while (wp != NULL && (wp->bottom_line+2 != window->top_line ||
                !wp->visible))
            wp = wp->next;
         if (wp != NULL)
            change = TRUE;
      }
      if (change != TRUE) {
         wp = g_status.window_list;
         while (wp != NULL && (wp->bottom_line != g_display.nlines ||
               !wp->visible))
            wp = wp->next;
         if (wp != NULL && wp != window)
            change = TRUE;
      }

      if (change == TRUE) {
         un_copy_line( window->cursor, window, TRUE );
         g_status.current_window = wp;
         g_status.current_file = wp->file_info;
      }
   }
}


/*
 * Name:    split_screen
 * Purpose: To split screen at the cursor.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   split the screen at the cursor position.
 */
void split_screen( windows *window )
{
register windows *wp;
register windows *win;  /* register pointer for window */
windows *temp;
file_infos *file;       /* file structure for file belonging to new window */

   win = window;
   /*
    * check that there is room for the window
    */
   if (win->bottom_line - win->cline < 2)
      error( WARNING, win->bottom_line, "move cursor up first" );
   else {
      file = win->file_info;
      if (create_window( &temp, win->cline+1, win->bottom_line,
                         file ) == ERROR) {
         error( WARNING, win->bottom_line, "out of memory" );
      } else {
         un_copy_line( win->cursor, win, TRUE );
         wp = temp;
         /*
          * record that the current window has lost some lines from
          *  the bottom for the new window, and adjust its page size
          *  etc accordingly.
          */
         win->bottom_line = win->cline;
         setup_window( win );
         display_current_window( win );

         /*
          * set up the new cursor position as appropriate
          */
         wp->rcol = win->rcol;
         wp->ccol = win->ccol;
         wp->bcol = win->bcol;
         wp->rline = win->rline;
         wp->cline = wp->cline + win->cline - win->top_line;
         if (wp->cline > wp->bottom_line)
            wp->cline = wp->bottom_line;
         wp->cursor = win->cursor;
         wp->visible = TRUE;
         wp->letter = file->next_letter++;

         /*
          * the new window becomes the current window.
          */
         g_status.current_window = wp;

         show_window_count( g_status.window_count );
         show_window_header( wp );
         display_current_window( wp );
      }
   }
}


/*
 * Name:    size_window
 * Purpose: To change the size of the current and one other window.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Use the Up and Down arrow keys to make the current window
 *          bigger or smaller.  The window above will either grow
 *          or contract accordingly.
 */
void size_window( windows *window )
{
char line_buff[(MAX_COLS+1)*2]; /* buffer for char and attribute  */
int func, c, above_change;
register windows *above;
register windows *win;
text_ptr p;
int resize;
char *instr =
       "Press Up or Down to change window size.  Press Return when done.";

   win = window;
   if (win->top_line != 1) {
      un_copy_line( win->cursor, win, TRUE );
      above_change = 0;
      save_screen_line( 0, win->bottom_line, line_buff );
      set_prompt( instr, win->bottom_line );

      /*
       * resizing only affects current window and above visible window
       */
      above = g_status.window_list;
      while (above->bottom_line + 2 != win->top_line || !above->visible)
         above = above->next;
      for (func=0; func != AbortCommand && func != Rturn; ) {

         /*
          * If user has redined the ESC and Return keys, make them Rturn and
          * AbortCommand in this function.
          */
         c = getkey( );
         func = getfunc( c );
         if (c == RTURN || func == NextLine || func == BegNextLine)
            func = Rturn;
         else if (c == ESC)
            func = AbortCommand;
         resize = FALSE;

         /*
          * if LineUp, make current window top line grow and bottom line
          * of above window shrink.   if window movement covers up current
          * line of window then we must adjust logical line and real line.
          */
         if (func == LineUp) {
            if (above->bottom_line > above->top_line) {
               if (win->rline == (win->cline - (win->top_line-1)))
                  --win->cline;
               --win->top_line;
               if (above->cline == above->bottom_line) {
                  above->cursor = cpb( above->cursor );
                  above->cursor = find_prev( above->cursor );
                  --above->rline;
                  --above->cline;
                  show_line_col( above );
                  --above_change;
               }
               --above->bottom_line;
               resize = TRUE;
               display_current_window( win );
               save_screen_line( 0, win->bottom_line, line_buff );
            }

         /*
          * if LineDown, make current window top line shrink and bottom line
          * of above window grow.   if window movement covers up current
          * line of window then we must adjust logical line and real line.
          */
         } else if (func == LineDown) {
            if (win->bottom_line > win->top_line) {
               if (win->cline == win->top_line) {
                  ++win->cline;
                  win->cursor = cpf( win->cursor );
                  p = find_next( win->cursor );
                  if (p != NULL) {
                     win->cursor = p;
                     ++win->rline;
                  }
               }
               ++win->top_line;
               ++above->bottom_line;
               display_current_window( above );
               ++above_change;
               resize = TRUE;
            }
         }

         /*
          * if we resize a window then update window size and current and
          * real lines if needed.
          */
         if (resize == TRUE) {
            setup_window( above );
            setup_window( win );
            show_window_header( win );
            set_prompt( instr, win->bottom_line );
         }
      }
      restore_screen_line( 0, win->bottom_line, line_buff );
   } else
     error( WARNING, win->bottom_line, "can not resize top window" );
}


/*
 * Name:    zoom_window
 * Purpose: To blow-up current window.
 * Date:    September 1, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Make all windows, visible and hidden, full size.
 */
void zoom_window( windows *window )
{
register windows *wp;

   if (window != NULL) {
      if (window->top_line != 1 || window->bottom_line != g_display.nlines) {
         un_copy_line( window->cursor, window, TRUE );
         for (wp=g_status.window_list; wp != NULL; wp=wp->next) {
            if (wp != window && wp->visible)
               wp->visible = FALSE;
            if (wp->top_line != 1)
               wp->cline = wp->cline - wp->top_line + 1;
            wp->top_line = 1;
            wp->bottom_line = g_display.nlines;
         }
         redraw_screen( window );
      }
   }
}


/*
 * Name:    next_hidden_window
 * Purpose: To display the window that is "behind" current window.
 * Date:    September 1, 1991
 * Passed:  window:   information allowing access to the current window
 */
void next_hidden_window( windows *window )
{
int poof = FALSE;
register windows *wp;

   if (window != NULL) {

      /*
       * look for next hidden window starting with current window.
       */
      wp = window;
      for (wp=window->next; wp != NULL && !poof; ) {
         if (!wp->visible)
            poof = TRUE;
         else
            wp = wp->next;
      }

      /*
       * if we haven't found an invisible window yet, start looking
       * for a hidden window from the beginning of the window list.
       */
      if (!poof) {
         for (wp=g_status.window_list; wp != NULL && !poof; ) {
            if (!wp->visible)
               poof = TRUE;
            else
               wp = wp->next;
         }
      }

      if (poof) {
         un_copy_line( window->cursor, window, TRUE );
         wp->cline = window->top_line + (wp->cline - wp->top_line);
         wp->top_line = window->top_line;
         wp->bottom_line = window->bottom_line;
         if (wp->cline > wp->bottom_line)
            wp->cline = wp->bottom_line;
         if ((wp->cline+1L - wp->top_line) > wp->rline)
            wp->cline = (int)wp->rline;
         wp->visible = TRUE;
         window->visible = FALSE;
         g_status.current_window = wp;
         redraw_current_window( wp );
      }
   }
}


/*
 * Name:    setup_window
 * Purpose: To set the page length and the center line of a window, based
 *           on the top and bottom lines.
 * Date:    June 5, 1991
 * Passed:  window: window to be set up
 */
void setup_window( windows *window )
{
   window->page = window->bottom_line - window->top_line -
                  g_status.overlap + 1;
   if (window->page < 1)
      window->page = 1;
}


/*
 * Name:    finish
 * Purpose: To remove the current window and terminate the program if no
 *           more windows are left.
 * Date:    June 5, 1991
 * Passed:  window: information allowing access to the current window
 * Notes:   Order of deciding which window becomes current window:
 *          1) If any invisible windows with same top and bottom line then
 *          first invisible one becomes current window.
 *          2) first available invisible window becomes current window.
 *          3) window above if it exists becomes current window
 *          4) window below if it exists becomes current window
 */
void finish( windows *window )
{
register windows *wp;   /* for scanning other windows */
register windows *win;  /* register pointer for window */
file_infos *file, *fp;  /* for scanning other files */
long number;            /* number of bytes removed / copied */
int poof;
int cline;
int top, bottom;
int max_letter;
int file_change = FALSE;

   win = window;
   un_copy_line( win->cursor, win, TRUE );
   /*
    * remove old window from list
    */
   top = 1;
   bottom = g_display.nlines;
   if (win->prev == NULL) {
      if (win->next == NULL)
         g_status.stop = TRUE;
      else
         g_status.window_list = win->next;
   } else
      win->prev->next = win->next;


   /*
    * remove all hidden windows that point to same file
    */
   if (win->next)
      win->next->prev = win->prev;
   file = win->file_info;
   for (wp=g_status.window_list; wp != NULL; wp=wp->next) {
      if (wp->file_info == file) {
         if (!wp->visible) {
            if (wp->prev == NULL) {
               if (wp->next == NULL)
                  g_status.stop = TRUE;
               else
                  g_status.window_list = wp->next;
            } else
               wp->prev->next = wp->next;
            if (wp->next)
               wp->next->prev = wp->prev;
            --wp->file_info->ref_count;
            free( wp );
            --g_status.window_count;
         }
      }
   }

   if (g_status.stop != TRUE) {
      /*
       * see if there are any invisible windows with same top and bottom
       * lines as this window.  start looking at end of window list.
       */
      poof = FALSE;
      wp = g_status.window_list;
      if (wp != NULL) {
         while (wp->next != NULL)
            wp = wp->next;
      }
      while (wp != NULL && poof == FALSE) {
         if (wp->top_line == win->top_line && !wp->visible) {
            poof = TRUE;
            top = win->top_line;
            bottom = win->bottom_line;
         } else
            wp = wp->prev;
      }

      if (poof == FALSE) {
         /*
          * see if there are any other invisible windows.  start looking
          * at the end of the window list.
          */
         wp = g_status.window_list;
         if (wp != NULL) {
            while (wp->next != NULL)
               wp = wp->next;
         }
         while (wp != NULL && poof == FALSE) {
            if (!wp->visible) {
               poof = TRUE;
               top = win->top_line;
               bottom = win->bottom_line;
            } else
               wp = wp->prev;
         }
         if (poof == FALSE) {
            /*
             * see if there are any windows above
             */
            wp = g_status.window_list;
            while (wp != NULL && poof == FALSE) {
               if (wp->bottom_line+2 == win->top_line) {
                  poof = TRUE;
                  top = wp->top_line;
                  bottom = win->bottom_line;
               } else
                  wp = wp->next;
            }
         }
         if (poof == FALSE) {
            /*
             * see if there are any windows below
             */
            wp = g_status.window_list;
            while (wp != NULL && poof == FALSE) {
               if (wp->top_line-2 == win->bottom_line) {
                  poof = TRUE;
                  top = win->top_line;
                  bottom = wp->bottom_line;
               } else
                  wp = wp->next;
            }
            if (wp == NULL)
               wp = g_status.window_list;
         }
      }
      wp->visible = TRUE;
      cline = wp->cline - wp->top_line;
      wp->top_line = top;
      wp->bottom_line = bottom;
      wp->cline = wp->top_line + cline;
      if (wp->cline > wp->bottom_line)
         wp->cline = wp->top_line;
      setup_window( wp );
      show_window_header( wp );

      /*
       * The window above, below, or previously invisible becomes the new
       * current window.
       */
      g_status.current_window = wp;
   }

   /*
    * free unused file memory if necessary
    */
   if (--file->ref_count == 0) {

      /*
       * if a block is marked, unmark it
       */
      if (file == g_status.marked_file) {
         g_status.marked      = FALSE;
         g_status.marked_file = NULL;
      }

      for (fp=g_status.file_list; fp != NULL; fp=fp->next) {
         if (fp->file_no > file->file_no)
            fp->file_no--;
      }
      file_change = file->file_no;

      /*
       * no window now refers to this file, so remove file from the list
       */
      if (file->prev == NULL)
         g_status.file_list = file->next;
      else
         file->prev->next = file->next;
      if (file->next)
         file->next->prev = file->prev;

      /*
       * close up the gap in the memory buffer
       */
      number = ptoul( g_status.end_mem ) - ptoul( file->end_text );
      hw_move( file->start_text, file->end_text, number );
      number = ptoul( file->end_text ) - ptoul( file->start_text );
      g_status.end_mem = addltop( -number, g_status.end_mem );
      addorsub_all_cursors( win, -number );
      adjust_start_end( file, -number );

      /*
       * free the memory taken by the file structure
       */
      free( file );
      if (--g_status.file_count) {
         show_file_count( g_status.file_count );
         show_avail_mem( );
      }
   }

   /*
    * free the memory taken by the window structure
    */
   free( win );
   --g_status.window_count;

   if (g_status.stop == FALSE) {
      g_status.current_file = wp->file_info;
      wp->file_info->dirty = LOCAL;
      show_window_count( g_status.window_count );
      if (file_change) {
         for (wp=g_status.window_list; wp!=NULL; wp=wp->next)
            if (wp->visible)
               show_window_number_letter( wp );
      } else {
         max_letter = 'a';
         for (wp=g_status.window_list; wp!=NULL; wp=wp->next) {
            if (wp->file_info == file && wp->letter > max_letter)
               max_letter = wp->letter;
         }
         if (max_letter < file->next_letter - 1)
            file->next_letter = max_letter + 1;
      }
   }
}


/*
 * Name:    create_window
 * Purpose: To allocate space for a new window structure and set up some
 *           of the relevant fields.
 * Date:    June 5, 1991
 * Passed:  window: pointer to window pointer
 *          top:    the top line of the new window
 *          bottom: the bottom line of the new window
 *          file:   the file structure to be associated with the new window
 * Returns: OK if window could be created
 *          ERROR if out of memory
 */
int  create_window( windows **window, int top, int bottom, file_infos *file )
{
windows *wp;            /* temporary variable - use it instead of **window */
register windows *prev;
int rc;                 /* return code */

   rc = OK;
   /*
    * allocate space for new window structure
    */
   if ((*window = (windows *)calloc( 1, sizeof(windows) )) == NULL) {
      error( WARNING, g_display.nlines, "out of memory for window" );
      rc = ERROR;
   } else {

     /*
      * set up appropriate fields
      */
      wp = *window;
      wp->file_info = file;
      wp->cline = wp->top_line = top+1;
      wp->bottom_line = bottom;
      wp->prev = NULL;
      wp->next = NULL;
      setup_window( wp );

      /*
       * add window into window list
       */
      prev = g_status.current_window;
      if (prev) {
         (*window)->prev = prev;
         if (prev->next)
            prev->next->prev = *window;
         (*window)->next = prev->next;
         prev->next = *window;
      }
      if (g_status.window_list == NULL)
         g_status.window_list = *window;

      /*
       * record that another window is referencing this file
       */
      ++file->ref_count;
      file->dirty = LOCAL;
      ++g_status.window_count;
   }
   return( rc );
}


/*
 * Name:    edit_file
 * Purpose: To allocate space for a new file structure and set up some
 *           of the relevant fields.
 * Date:    June 5, 1991
 * Passed:  name:  name of the file to edit
 * Returns: OK if file structure could be created
 *          ERROR if out of memory
 */
int  edit_file( char *name )
{
int rc;           /* return code */
int existing;
int line;
register file_infos *file; /* file structure for file belonging to new window */
file_infos *fp;
text_ptr next;      /* text pointer */
long size;

   rc = OK;
   line = g_display.nlines;
   if (hw_fattrib( name ) == OK) {
      existing = TRUE;
      /*
       * g_status.temp_end is set last character read in file
       */
      rc = load_file( name );
   } else {
      existing = FALSE;
      g_status.temp_end = cpf( g_status.end_mem );
   }

   if (rc != ERROR) {
      /*
       * allocate a file structure for the new file
       */
      file = (file_infos *)calloc( 1, sizeof(file_infos) );
      if (file == NULL) {
         error( WARNING, g_display.nlines, "out of memory for file info" );
         rc = ERROR;
      } else {
         /*
          * add file into list
          */
         file->prev = NULL;
         file->next = NULL;
         if (g_status.file_list == NULL)
            g_status.file_list = file;
         else {
            fp = g_status.current_file;
            file->prev = fp;
            if (fp->next)
               fp->next->prev = file;
            file->next = fp->next;
            fp->next = file;
         }
      }
      if (rc != ERROR) {
         /*
          * set up all the info we need to know about a file, and
          *  record that we have used some more memory.
          */
         strcpy( file->file_name, name );
         get_fattr( name, &file->file_attrib );
         file->start_text = g_status.end_mem;
         *g_status.temp_end = CONTROL_Z;
         g_status.temp_end = cpf( g_status.temp_end );
         g_status.end_mem = g_status.temp_end + 1;
         g_status.temp_end = g_status.end_mem;
         file->end_text = g_status.end_mem;
         size = 0l;
         if (existing) {
            next = file->start_text = cpf( file->start_text );
            while ((next = find_next( next )) != NULL)
               size++;
         }
         file->length      = size;
         file->block_end   = file->block_start = NULL;
         file->block_bc    = file->block_ec = 0;
         file->block_br    = file->block_er = 0l;
         file->block_type  = NOTMARKED;
         file->modified    = FALSE;
         file->new_file    = !existing;
         file->ref_count   = 0;
         file->next_letter = 'a';
         file->file_no     = ++g_status.file_count;
         g_status.current_file = file;
      }
   }
   return( rc );
}


/*
 * Name:    edit_another_file
 * Purpose: Bring in another file to editor.
 * Date:    June 5, 1991
 * Passed:  window: information allowing access to the current window
 * Notes:   New window replaces old window.  Old window becomes invisible.
 */
void edit_another_file( windows *window )
{
char name[MAX_COLS];              /* new name for file */
register int ok;
register windows *win;           /* put window pointer in a register */

   win = window;
   un_copy_line( win->cursor, win, TRUE );
   /*
    * read in name, no default
    */
   name[0] = '\0';
   if (get_name( "File to edit: ", win->bottom_line, name,
                 g_display.message_color ) == OK) {
      ok = edit_file( name );
      if (ok != ERROR) {
         ok = initialize_window( );
         if (ok != ERROR) {
            win = g_status.current_window;
            redraw_current_window( win );
            show_file_count( g_status.file_count );
            show_window_count( g_status.window_count );
            show_avail_mem( );
         }
      }
   }
}


/*
 * Name:    edit_next_file
 * Purpose: edit next file on command line.
 * Date:    January 6, 1992
 * Passed:  window: information allowing access to the current window
 * Notes:   New window replaces old window.  Old window becomes invisible.
 */
int  edit_next_file( windows *window )
{
char name[MAX_COLS];            /* new name for file */
int i;
register int rc = ERROR;
register windows *win;           /* put window pointer in a register */

   if (g_status.arg < g_status.argc) {
      win = window;
      if (win != NULL)
         un_copy_line( win->cursor, win, TRUE );

      /*
       * while we haven't found a valid file, search thru the command
       * line path.
       * we may have an invalid file name when we finish matching all
       * files according to a pattern.  then, we need to go to the next
       * command line argument if it exists.
       */
      while (rc == ERROR && g_status.arg < g_status.argc) {

         /*
          * if we haven't starting searching for a file, check to see if
          * the file is a valid file name.  if no file is found, then let's
          * see if we can find according to a search pattern.
          */
         if (g_status.found_first == FALSE) {
            strcpy( name, g_status.argv[g_status.arg] );
            rc = get_fattr( name, &i );

            /*
             * a new or blank file generates a return code of 2.
             * a pattern with wild cards generates a return code of 3.
             */
            if (rc == OK || rc == 2) {
               ++g_status.arg;
               rc = OK;

            /*
             * if we get this far, we got an invalid path name.
             *  let's try to find a matching file name using pattern.
             */
            } else {
               rc = findfirst( &g_status.dta, name );

               /*
                * if we found a file using wildcard characters,
                * set the g_status.found_first flag to true so we can
                * find the next matching file.  we need to save the
                * pathname stem so we know which directory we are working in.
                */
               if (rc == OK) {
                  g_status.found_first = TRUE;
                  i = strlen( name ) - 1;
                  while (i >= 0) {
                     if (name[i] == ':' || name[i] == '\\')
                        break;
                     --i;
                  }
                  name[++i] = '\0';
                  strcpy( g_status.path, name );
                  strcpy( name, g_status.path );
                  strcat( name, g_status.dta.name );
               } else {
                  ++g_status.arg;
                  if (win != NULL)
                     error( WARNING, win->bottom_line,
                            "Invalid path or file name " );
               }
            }
         } else {

            /*
             * we already found one file with wild card characters,
             * find the next matching file.
             */
            rc = findnext( &g_status.dta );
            if (rc == OK) {
               strcpy( name, g_status.path );
               strcat( name, g_status.dta.name );
            } else {
               g_status.found_first = FALSE;
               ++g_status.arg;
            }
         }

         /*
          * if everything is everything so far, set up the file
          * and window structures and bring the file into the editor.
          */
         if (rc == OK) {
            rc = edit_file( name );
            if (rc != ERROR) {
               rc = initialize_window( );
               if (rc != ERROR && win != NULL) {
                  win = g_status.current_window;
                  redraw_current_window( win );
                  show_file_count( g_status.file_count );
                  show_window_count( g_status.window_count );
                  show_avail_mem( );
               }
            }

         /*
          * either there are no more matching files or we had an
          * invalid file name, set rc to ERROR and let's look at the
          * next file name or pattern on the command line.
          */
         } else
            rc = ERROR;
      }
   }
   if (rc == ERROR  &&  g_status.arg >= g_status.argc  &&  win != NULL)
      error( WARNING, win->bottom_line, "No more files to load " );
   return( rc );
}
